:- module(abstract, [abstract/3]).
:- use_module(library(lists)).
:- use_module(library(terms)).

:- dynamic dis_size/1, abs_size/1, con_size/1.

abstract(Stream, Type, Abs_Size) :-
	findall(op(Level, Assoc, Op), current_op(Level, Assoc, Op), Old_Ops),
	retractall(fixpoint:head_to_clause(_, _)),
 	retractall(fixpoint:atom_to_clause(_, _, _, _)),
 	retractall(fixpoint:atom_to_dynamic(_)),
 	retractall(fixpoint:atom_to_dynamic(_, _, _, _)),
	abolish(dis_size, 1),
	abolish(abs_size, 1),
	abolish(con_size, 1),
	assert(dis_size(1)),
	assert(abs_size(0)),
	assert(con_size(0)),
    	read_clauses(Stream, Cs),
%	trace, leash([exception]),
	abs_clauses(Cs, Type),
%	halt,
	retract(abs_size(Abs_Size)),
	retract(con_size(Con_Size)),
	format("(~w/~w clauses) ", [Abs_Size, Con_Size]),
	ttyflush,
	findall(op(Level, Assoc, Op), current_op(Level, Assoc, Op), New_Ops),
	remove_ops(New_Ops),
	assert_ops(Old_Ops).

remove_ops([]).
remove_ops([op(_, Assoc, Op) | Ops]) :-
	op(0, Assoc, Op),
	remove_ops(Ops).

assert_ops([]).
assert_ops([op(Level, Assoc, Op) | Ops]) :-
	op(Level, Assoc, Op),
	assert_ops(Ops).

read_clauses(Stream, Cs) :-
   	read(Stream, Term),	
	expand_term(Term, Clause),
    	read_clause(Clause, Stream, Cs).

read_clause(end_of_file, _, []) :-
	!.
read_clause((:- op(A, B, C)), Stream, Cs) :- 
	!,
    	op(A, B, C),
	read_clauses(Stream, Cs).
read_clause((:- _), Stream, Cs) :- 
	!,
	read_clauses(Stream, Cs).
read_clause((H :- B), Stream, Cs) :- 
	!,
	Cs = [(H :- B) | Rest_Cs],
	read_clauses(Stream, Rest_Cs).
read_clause(H, Stream, Cs) :- 
	Cs = [(H :- true) | Rest_Cs],
	read_clauses(Stream, Rest_Cs).

abs_clauses([], _).
abs_clauses([C | Cs_In], Type) :-
	abs_clause(C, Type, Cs_Out, Cs_In),
	abs_clauses(Cs_Out, Type).

abs_clause((H :- B), Type, Cs_Out, Cs_In) :-
	H =.. [Name | HArgs],
	simplify_args(HArgs, RArgs, Type, Gs, Bs),
	R =.. [Name | RArgs],
    	abs_goal(B, Type, Bs, [], Cs_Out, Cs_In),
    	cond_assert_clause(R, Gs).

cond_assert_clause(Head, Body) :-
	retract(con_size(Con_Size)),
        Con_Size1 is Con_Size + 1,
        assert(con_size(Con_Size1)),
	in_database(Head, Body),
	!.
cond_assert_clause(Head, Body) :-
	retract(abs_size(Abs_Size)),
        Abs_Size1 is Abs_Size + 1,
        assert(abs_size(Abs_Size1)),
%	unflatten_body(Body, Goal),
%	portray_clause((Head :- Goal)),
	asserta(fixpoint:head_to_clause(Head, Body)),
        atoms_assert(Body, Head, []).

unflatten_body([], true).
unflatten_body([Atom], Goal) :-
	!,
	Goal = Atom.
unflatten_body([Atom | Atoms], (Atom, Goal)) :-
	unflatten_body(Atoms, Goal).

atoms_assert([], _, _).
atoms_assert([Atom | Atoms], Head, Rev_Left) :- 
        atom_assert(Atom, Atoms, Head, Rev_Left).

atom_assert(fail, Atoms, Head, Rev_Left) :- 
	!,
	atoms_assert(Atoms, Head, [fail | Rev_Left]).
atom_assert(Atom, Atoms, Head, Rev_Left) :- 
	functor(Atom, if, _),
	!,
	atoms_assert(Atoms, Head, [Atom | Rev_Left]).
atom_assert(Atom, Atoms, Head, Rev_Left) :- 
	functor(Atom, iff, _),
	!,
	atoms_assert(Atoms, Head, [Atom | Rev_Left]).
atom_assert(Atom, Atoms, Head, Rev_Left) :- 
	reverse(Rev_Left, Left),
 	asserta(fixpoint:atom_to_clause(Atom, Head, Left, Atoms)),
	atoms_assert(Atoms, Head, [Atom | Rev_Left]).

in_database(Head, Body) :-
	functor(Head, Name, Arity),
	functor(Copy_Head, Name, Arity),
	fixpoint:head_to_clause(Copy_Head, Copy_Body),
	variant((Head, Body), (Copy_Head, Copy_Body)),
	!.

abs_goal(G, Type, Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	var(G),
	!,
	abs_goal(call(G), Type, Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_goal((P; Q), _, Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	Gs_Out = [G | Gs_In], 
	term_variables((P; Q), Args),
	fresh_name(Name),
	G =.. [Name | Args],
	Cs_Out = [(G :- P), (G :- Q) | Cs_In].
abs_goal((P -> Q), Type, Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	abs_goal((P, Q), Type, Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_goal(\+(G), Type, Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	abs_goal(not(G), Type, Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_goal((P, Q), Type, Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	abs_goal(P, Type, Gs_Out, Gs_Tmp, Cs_Out, Cs_Tmp),
	abs_goal(Q, Type, Gs_Tmp, Gs_In, Cs_Tmp, Cs_In).
abs_goal(!, _, Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	Gs_Out = Gs_In,
	Cs_Out = Cs_In.
abs_goal(not(G), Type, Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	abs_goal(((G, !, fail); true), Type, Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_goal(G, Type, Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	abs_builtin(G, Type, Gs_Out, Gs_In, Cs_Out, Cs_In),
	!.
abs_goal(P, Type, Gs_Out, Gs_In, Cs, Cs) :-
	P =.. [Name | PArgs],
	simplify_args(PArgs, RArgs, Type, Gs_Out, [R | Gs_In]),
	R =.. [Name | RArgs].

fresh_name(Name)  :-
	retract(dis_size(Size)),
	Size1 is Size + 1,
	asserta(dis_size(Size1)),
	name(Size, Size_String),
	name(Name, [95, 95 | Size_String]).

abs_ground_if(Term1, Term2, Gs_Out, Gs_In) :-
	term_variables(Term1, Vars1),
	term_variables(Term2, Vars2),
	abs_if_vars(Vars1, Vars2, Gs_Out, Gs_In).

abs_if_vars([Var1], Vars2, Gs_Out, Gs_In) :-
	!,
	abs_if(Var1, Vars2, Gs_Out, Gs_In).
abs_if_vars(Vars1, Vars2, Gs_Out, Gs_In) :-
	abs_iff(Var, Vars1, Gs_Out, Gs_Tmp),
	abs_if(Var, Vars2, Gs_Tmp, Gs_In).

abs_ground_iff(Term1, Term2, Gs_Out, Gs_In) :-
	term_variables(Term1, Vars1),
	term_variables(Term2, Vars2),
	abs_ground_iff_vars(Vars1, Vars2, Gs_Out, Gs_In).

abs_ground_iff_vars([Var1], Vars2, Gs_Out, Gs_In) :-
	!,
	abs_iff(Var1, Vars2, Gs_Out, Gs_In).
abs_ground_iff_vars(Vars1, [Var2], Gs_Out, Gs_In) :-
	!,
	abs_iff(Var2, Vars1, Gs_Out, Gs_In).
abs_ground_iff_vars(Vars1, Vars2, Gs_Out, Gs_In) :-
	abs_iff(Var, Vars1, Gs_Out, Gs_Tmp),
	abs_iff(Var, Vars2, Gs_Tmp, Gs_In).

abs_if(Left_Arg, [Right_Arg], Gs_Out, Gs_In) :-
	!,
	abs_iff(Right_Arg, [Left_Arg, _], Gs_Out, Gs_In).
abs_if(Left_Arg, Right_Args, Gs_Out, Gs_In) :-
	abs_iff(Var, Right_Args, Gs_Out, Gs_Tmp),
	abs_iff(Var, [Left_Arg, _], Gs_Tmp, Gs_In).

abs_iff(Left_Arg, Right_Args, [G | Gs], Gs) :-
	G =.. [iff, Left_Arg | Right_Args]. 

abs_ground_term(Term, Gs_Out, Gs_In) :-
	term_variables(Term, Vars),
	abs_herbrand_ground_vars(Vars, Gs_Out, Gs_In).

abs_herbrand_ground_vars([], Gs, Gs).
abs_herbrand_ground_vars([Var | Vars], Gs_Out, Gs_In) :-
	abs_iff(Var, [], Gs_Out, Gs_Tmp),
	abs_herbrand_ground_vars(Vars, Gs_Tmp, Gs_In).

abs_herbrand_constraint(LTerm, RTerm, Gs_Out, Gs_In) :-
	var(LTerm),
	!,
	term_variables(RTerm, Vars),
	abs_iff(LTerm, Vars, Gs_Out, Gs_In).
abs_herbrand_constraint(LTerm, RTerm, Gs_Out, Gs_In) :-
	var(RTerm),
	!,
	term_variables(LTerm, Vars),
	abs_iff(RTerm, Vars, Gs_Out, Gs_In).
abs_herbrand_constraint(LTerm, RTerm, Gs_Out, Gs_In) :-
	LTerm =.. [Name | LArgs],
	RTerm =.. [Name | RArgs],
	abs_herbrand_constraints(LArgs, RArgs, Gs_Out, Gs_In),
	!.
abs_herbrand_constraint(_, _, [fail | Gs], Gs).

abs_herbrand_constraints([], [], Gs, Gs).
abs_herbrand_constraints([LArg | LArgs], [RArg | RArgs], Gs_Out, Gs_In) :-
	abs_herbrand_constraint(LArg, RArg, Gs_Out, Gs_Tmp),
	abs_herbrand_constraints(LArgs, RArgs, Gs_Tmp, Gs_In).

simplify_args(PArgs, RArgs, Type, Gs_Out, Gs_In) :-
	simplify_args(PArgs, RArgs, Type, [], _, Gs_Out, Gs_In).

simplify_args([], [], _, OArgs, OArgs, Gs, Gs).
simplify_args([PArg | PArgs], [RArg | RArgs], Type, OArgs_In, OArgs_Out, Gs_Out, Gs_In) :-
	term_variables(PArg, PVars),
	simplify_arg(Type, PVars, PArg, RArg, OArgs_In, OArgs_Mid, Gs_Out, Gs_Tmp),
	simplify_args(PArgs, RArgs, Type, OArgs_Mid, OArgs_Out, Gs_Tmp, Gs_In).

simplify_arg(herbrand, PVars, PArg, RArg, OArgs_In, OArgs_Mid, Gs_Out, Gs_Tmp) :-
	simplify_herbrand_arg(PVars, PArg, RArg, OArgs_In, OArgs_Mid, Gs_Out, Gs_Tmp).
simplify_arg(numeric, PVars, PArg, RArg, OArgs_In, OArgs_Mid, Gs_Out, Gs_Tmp) :-
	simplify_numeric_arg(PVars, PArg, RArg, OArgs_In, OArgs_Mid, Gs_Out, Gs_Tmp).

simplify_herbrand_arg([], _, RArg, OArgs, OArgs, Gs_Out, Gs_In) :-
	!,
	abs_iff(RArg, [], Gs_Out, Gs_In).
simplify_herbrand_arg([PVar], _, RArg, OArgs_In, OArgs_Out, Gs_Out, Gs_In) :-
	!,
	simplify_old_arg(PVar, RArg, OArgs_In, OArgs_Out, Gs_Out, Gs_In).
simplify_herbrand_arg(PVars, _, RArg, OArgs, OArgs, Gs_Out, Gs_In) :-
	abs_iff(RArg, PVars, Gs_Out, Gs_In).

simplify_old_arg(PVar, RArg, OArgs_In, OArgs_Out, Gs_Out, Gs_In) :-
	my_memberchk(OArgs_In, PVar),
	!,
	OArgs_Out = OArgs_In,
	abs_iff(PVar, [RArg], Gs_Out, Gs_In).
simplify_old_arg(Arg, Arg, OArgs, [Arg | OArgs], Gs, Gs).

abs_builtin(Atom, herbrand, Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	abs_herbrand_table(Atom, Gs_Out, Gs_In, Cs_Out, Cs_In),
	!.
abs_builtin(Atom, herbrand, Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	predicate_property(Atom, built_in),
	!,
	Gs_Out = Gs_In,
	Cs_Out = Cs_In,
	format("~n{warning, ignoring ~w}", [Atom]),
	true.
abs_builtin(Atom, numeric, Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	abs_numeric_table(Atom, Gs_Out, Gs_In, Cs_Out, Cs_In).

abs_herbrand_table(Goal, Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	functor(Goal, Name, Arity),
	abs_herbrand_ground_table(Name, Arity),
	!,
	Cs_Out = Cs_In,
	abs_ground_term(Goal, Gs_Out, Gs_In).
abs_herbrand_table(Goal, Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	functor(Goal, Name, Arity),
	abs_herbrand_nonground_table(Name, Arity),
	!,
	Cs_Out = Cs_In,
	Gs_Out = Gs_In.
abs_herbrand_table(Goal, Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	functor(Goal, Name, Arity),
	abs_herbrand_fail_table(Name, Arity),
	!,
	Cs_Out = Cs_In,
	Gs_Out = [fail | Gs_In].
abs_herbrand_table((Term1 = Term2), Gs_Out, Gs_In, Cs, Cs) :-
	abs_herbrand_constraint(Term1, Term2, Gs_Out, Gs_In).
abs_herbrand_table((Term1 =.. Term2), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_iff(Term1, Term2, Gs_Out, Gs_In).
abs_herbrand_table((Term1 =.. Term2), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_iff(Term1, Term2, Gs_Out, Gs_In).
abs_herbrand_table((Term1 == Term2), Gs_Out, Gs_In, Cs, Cs) :-
	abs_herbrand_constraint(Term1, Term2, Gs_Out, Gs_In).
abs_herbrand_table('C'(Term1, Term2, Term3), Gs_Out, Gs_In, Cs, Cs) :-
	abs_herbrand_constraint(Term1, [Term2 | Term3], Gs_Out, Gs_In).
abs_herbrand_table(arg(Index, Term, Arg), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Index, Gs_Out, Gs_Tmp),
	abs_ground_if(Arg, Term, Gs_Tmp, Gs_In).
abs_herbrand_table(assert(Clause), Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	abs_assert(Clause, Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_herbrand_table(asserta(Clause), Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	abs_herbrand_table(assert(Clause), Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_herbrand_table(assertz(Clause), Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	abs_herbrand_table(assert(Clause), Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_herbrand_table(bagof(Target, G, Bag), Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	abs_bagof(Target, G, Bag, Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_herbrand_table(call(G), Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	abs_call(G, Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_herbrand_table(copy_term(Term1, Term2), Gs_Out, Gs_In, Cs, Cs) :-
	abs_copy_term(Term1, Term2, Gs_Out, Gs_In).
abs_herbrand_table(compare(Term, _, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(current_predicate(Term, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(findall(Target, G, Bag), Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	abs_findall(Target, G, Bag, Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_herbrand_table(format(Term, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(format(Term1, Term2, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term([Term1, Term2], Gs_Out, Gs_In).
abs_herbrand_table(functor(_, Term1, Term2), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term([Term1, Term2], Gs_Out, Gs_In).
abs_herbrand_table(hash_term(_, Term), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(hash_term(_, Term1, Term2, Term3), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term([Term1, Term2, Term3], Gs_Out, Gs_In).
abs_herbrand_table(instance(Term, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(keysort(Term1, Term2), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_iff(Term1, Term2, Gs_Out, Gs_In).
abs_herbrand_table(length(_, Term), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(portray_clause(Term, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(predicate_property(_, Term), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(print(Term, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(read(Term, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(retract(Clause), Gs_Out, Gs_In, Cs, Cs) :-
	abs_retract(Clause, Gs_Out, Gs_In).
abs_herbrand_table(setof(Target, G, Bag), Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	abs_herbrand_table(bagof(Target, G, Bag), Gs_Out, Gs_In, Cs_Out, Cs_In).
abs_herbrand_table(sort(Term1, Term2), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_iff(Term1, Term2, Gs_Out, Gs_In).
abs_herbrand_table(source_file(_, Term), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(write(Term, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(write_canonical(Term, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).
abs_herbrand_table(writeq(Term, _), Gs_Out, Gs_In, Cs, Cs) :-
	abs_ground_term(Term, Gs_Out, Gs_In).

abs_herbrand_ground_table(>, 2).
abs_herbrand_ground_table(<, 2).
abs_herbrand_ground_table(=<, 2).
abs_herbrand_ground_table(>=, 2).
abs_herbrand_ground_table(=:=, 2).
abs_herbrand_ground_table(=\=, 2).
abs_herbrand_ground_table(abolish, 1).
abs_herbrand_ground_table(abolish, 2). 
abs_herbrand_ground_table(absolute_file_name, 2).
abs_herbrand_ground_table(atom, 1).
abs_herbrand_ground_table(atom_chars, 2).
abs_herbrand_ground_table(atomic, 1).
abs_herbrand_ground_table(compile, 1).
abs_herbrand_ground_table(consult, 1).
abs_herbrand_ground_table(character_count, 2).
abs_herbrand_ground_table(close, 1).
abs_herbrand_ground_table(current_atom, 1).
abs_herbrand_ground_table(current_input, 1). 
abs_herbrand_ground_table(current_module, 1).
abs_herbrand_ground_table(current_module, 2).
abs_herbrand_ground_table(current_op, 3).
abs_herbrand_ground_table(current_output, 1). 
abs_herbrand_ground_table(current_stream, 3).
abs_herbrand_ground_table(ensure_loaded, 1).
abs_herbrand_ground_table(erase, 1).
abs_herbrand_ground_table(float, 1).
abs_herbrand_ground_table(flush_output, 1).
abs_herbrand_ground_table(get, 1). 
abs_herbrand_ground_table(get, 2).
abs_herbrand_ground_table(get0, 1).
abs_herbrand_ground_table(get0, 2).
abs_herbrand_ground_table(ground, 1).
abs_herbrand_ground_table(integer, 1).
abs_herbrand_ground_table(is, 2).
abs_herbrand_ground_table(leash, 1).
abs_herbrand_ground_table(line_count, 2). 
abs_herbrand_ground_table(line_position, 2).
abs_herbrand_ground_table(load, 1). 
abs_herbrand_ground_table(name, 2).
abs_herbrand_ground_table(nl, 1).
abs_herbrand_ground_table(number, 1).
abs_herbrand_ground_table(number_chars, 2).
abs_herbrand_ground_table(numbervars, 3).
abs_herbrand_ground_table(op, 3).
abs_herbrand_ground_table(open, 3). 
abs_herbrand_ground_table(open, 4).
abs_herbrand_ground_table(open_null_stream, 1). 
abs_herbrand_ground_table(peek_char, 1).
abs_herbrand_ground_table(peek_char, 2).
abs_herbrand_ground_table(prolog_flag, 2). 
abs_herbrand_ground_table(prolog_flag, 3).
abs_herbrand_ground_table(prompt, 2).
abs_herbrand_ground_table(put, 1).
abs_herbrand_ground_table(put, 2).
abs_herbrand_ground_table(reconsult, 1).
abs_herbrand_ground_table(see, 1). 
abs_herbrand_ground_table(seeing, 1).
abs_herbrand_ground_table(set_input, 1).
abs_herbrand_ground_table(set_output, 1).
abs_herbrand_ground_table(skip, 1).
abs_herbrand_ground_table(skip, 2).
abs_herbrand_ground_table(skip_line, 1).
abs_herbrand_ground_table(source_file, 1). 
abs_herbrand_ground_table(statistics, 2).
abs_herbrand_ground_table(stream_code, 2).
abs_herbrand_ground_table(tab, 1).
abs_herbrand_ground_table(tab, 2).
abs_herbrand_ground_table(tell, 1).
abs_herbrand_ground_table(telling, 1).
abs_herbrand_ground_table(ttyget, 1).
abs_herbrand_ground_table(ttyget0, 1).
abs_herbrand_ground_table(ttyput, 1).
abs_herbrand_ground_table(ttyskip, 1).
abs_herbrand_ground_table(ttytab, 1).
abs_herbrand_ground_table(use_module, 2).
abs_herbrand_ground_table(use_module, 3).
abs_herbrand_ground_table(version, 1).

abs_herbrand_nonground_table(@<, 2).
abs_herbrand_nonground_table(@>, 2).
abs_herbrand_nonground_table(@=<, 2).
abs_herbrand_nonground_table(@>=, 2).
abs_herbrand_nonground_table(==, 2).
abs_herbrand_nonground_table(\==, 2).
abs_herbrand_nonground_table(break, 0).
abs_herbrand_nonground_table(callable, 1).
abs_herbrand_nonground_table(compound, 1). 
abs_herbrand_nonground_table(debug, 0).
abs_herbrand_nonground_table(debugging, 0).
abs_herbrand_nonground_table(dif, 2).
abs_herbrand_nonground_table(display, 1).
abs_herbrand_nonground_table(expand_term, 2). 
abs_herbrand_nonground_table(fileerrors, 0). 
abs_herbrand_nonground_table(garbage_collect, 0).
abs_herbrand_nonground_table(gc, 0).
abs_herbrand_nonground_table(help, 0).
abs_herbrand_nonground_table(listing, 0).
abs_herbrand_nonground_table(listing, 1).
abs_herbrand_nonground_table(nl, 0). 
abs_herbrand_nonground_table(nodebug, 0). 
abs_herbrand_nonground_table(nofileerrors, 0).
abs_herbrand_nonground_table(nogc, 0).
abs_herbrand_nonground_table(nonvar, 1).
abs_herbrand_nonground_table(nospyall, 0).
abs_herbrand_nonground_table(notrace, 0). 
abs_herbrand_nonground_table(otherwise, 0).
abs_herbrand_nonground_table(phrase, 2).
abs_herbrand_nonground_table(phrase, 3).
abs_herbrand_nonground_table(print, 1).
abs_herbrand_nonground_table(read, 1).
abs_herbrand_nonground_table(repeat, 0).
abs_herbrand_nonground_table(retractall, 1).
abs_herbrand_nonground_table(subsumes_chk, 2).
abs_herbrand_nonground_table(seen, 0).
abs_herbrand_nonground_table(simple, 1).
abs_herbrand_nonground_table(skip_line, 0).
abs_herbrand_nonground_table(statistics, 0).
abs_herbrand_nonground_table(system, 1). % pre-history
abs_herbrand_nonground_table(told, 0).
abs_herbrand_nonground_table(true, 0).
abs_herbrand_nonground_table(ttyflush, 0).
abs_herbrand_nonground_table(ttynl, 0).
abs_herbrand_nonground_table(var, 1).
abs_herbrand_nonground_table(version, 0).
abs_herbrand_nonground_table(write, 1).
abs_herbrand_nonground_table(writeq, 1). 
abs_herbrand_nonground_table(write_canonical, 1).

abs_herbrand_fail_table(abort, 0).
abs_herbrand_fail_table(fail, 0).
abs_herbrand_fail_table(false, 0).
abs_herbrand_fail_table(halt, 0).
abs_herbrand_fail_table(halt, 1).

my_difference([], _, []).
my_difference([Arg | Args], Vars, Delta) :-
	my_memberchk(Vars, Arg),
	!,
	my_difference(Args, Vars, Delta).
my_difference([Arg | Args], Vars, [Arg | Delta]) :-
	my_difference(Args, Vars, Delta).

my_memberchk([Arg | _], Var) :-
	Arg == Var,
	!.
my_memberchk([_ | Args], Var) :-
	my_memberchk(Args, Var).
