:- module(calibrate, [calibrate/2]).
:- use_module(library(terms)).

:- dynamic msg_atom/2.

calibrate(con_nonground_model, Count) :- 	
	count_deter_atoms(Count).
calibrate(det_nonground_model, Count) :- 	
	count_deter_atoms(Count).
calibrate(msg_nonground_model, Count) :- 	
	count_deter_atoms(Count).
calibrate(pos_nonground_model, Count) :- 	
	count_pos_atoms(Count).

count_pos_atoms(Count) :-
	retractall(msg_atom(_, _)),
	msg_pos_atoms,
	count_msg_atoms(0, Count).

msg_pos_atoms :-
	retract(fixpoint:user_atom(Atom, Tag)),
	msg_pos_atom(Atom, Tag),
	fail.
msg_pos_atoms.

msg_pos_atom(New_Atom, Tag) :-   
	functor(New_Atom, Name, Arity),
	functor(Old_Atom, Name, Arity),
	retract(msg_atom(Old_Atom, Tag)),
	!,
	term_subsumer(Old_Atom, New_Atom, Msg_Atom),
	assert(msg_atom(Msg_Atom, Tag)).
msg_pos_atom(Atom, Tag) :-   
	assert(msg_atom(Atom, Tag)).

count_deter_atoms(Count) :-
	count_deter_atoms(0, Count).

count_deter_atoms(Count_Acc, Count_Out) :-
        retract(fixpoint:user_atom(Atom, _)), 
	!,
	Atom =.. [_ | Args],
	count_args(Args, Count_Acc, Count_Tmp),
        count_deter_atoms(Count_Tmp, Count_Out).
count_deter_atoms(Count, Count).

count_share_atoms(Count) :-
	count_share_atoms(0, Count).

count_share_atoms(Count_Acc, Count_Out) :-
        retract(fixpoint:user_atom(Atom, _, Share, _Iterate)), 
	!,
%	format("~n~w", [Iterate]),
	term_variables(Share, Vars),
	length(Vars, Share_Vars),
	functor(Atom, _, Total_Vars),
	Count_Tmp is Count_Acc + Total_Vars - Share_Vars,
        count_share_atoms(Count_Tmp, Count_Out).
count_share_atoms(Count, Count).

count_msg_atoms(Count_Acc, Count_Out) :-
        retract(msg_atom(Atom, _)), 
	!,
	Atom =.. [_ | Args],
	count_args(Args, Count_Acc, Count_Tmp),
        count_msg_atoms(Count_Tmp, Count_Out).
count_msg_atoms(Count, Count).

count_args([], C, C).
count_args([Arg | Args], CI, CO) :-
	Arg == true,
	!,
	KI is CI + 1,
	count_args(Args, KI, CO).
count_args([_ | Args], CI, CO) :-
	count_args(Args, CI, CO).
