:- module(deter_fixpoint, [td_deter_fixpoint/1]).
:- use_module(library(terms)).
:- use_module(library(lists)).

td_deter_fixpoint(Goal) :-
	retractall(fixpoint:user_atom(_, _)),
%	spy(deter_iffs),
	call_cond_assert(Goal, Goal),
%       profile_reset(_),
       	(call_solve(Goal); true),
%	profile_data(_, execution_time, predicate, Data),
%      	format("~nprofile = ~w~n", [Data]),
	true.

call_solve(Head) :-
        fixpoint:head_to_clause(Head, Body), 
        solve_right(Body, [], Iffs), 
	deter_iffs(true, Iffs, _),
        answ_cond_assert(Head, New_Head), 
        answ_solve(New_Head).

answ_solve(Atom) :-
        fixpoint:atom_to_clause(Atom, Head, Left, Right),
        fixpoint:user_atom(Head, call), 
        solve_left(Left, [], Iffs_Tmp), 
        solve_right(Right, Iffs_Tmp, Iffs_Out), 
	deter_iffs(true, Iffs_Out, _),
        answ_cond_assert(Head, New_Head), 
        answ_solve(New_Head).

solve_left([], Iffs, Iffs).
solve_left([Atom | Atoms], Iffs_In, Iffs_Out) :- 
	fixpoint:user_atom(Atom, answ), 
	!,
	solve_left(Atoms, Iffs_In, Iffs_Out).
solve_left([Atom | Atoms], Iffs_In, Iffs_Out) :- 
	Atom =.. [Name, Arg | Args],
	New_Atom =.. [Name, Arg, Args],
	solve_left(Atoms, [New_Atom | Iffs_In], Iffs_Out).

solve_right([], Iffs, Iffs).
solve_right([Atom | Atoms], Iffs_In, Iffs_Out) :- 
	Atom =.. [iff, Arg | Args],
	!,
	solve_right(Atoms, [iff(Arg, Args) | Iffs_In], Iffs_Out).
solve_right([fail | _], _, _) :- 
	!,
	fail.
solve_right([Atom | Atoms], Iffs_In, Iffs_Out) :-
	deter_iffs(true, Iffs_In, Iffs_Tmp),
	solve_right(Atoms, Iffs_Tmp, Iffs_Out, Atom).

solve_right(Atoms, Iffs_In, Iffs_Out, Atom) :-
        fixpoint:user_atom(Atom, answ),
	solve_right(Atoms, Iffs_In, Iffs_Out).
solve_right(_, _, _, Atom) :-
        call_cond_assert(Atom, New_Atom),	
	call_solve(New_Atom).

order_iffs([], Accs, Iffs_Out) :-
	keysort(Accs, Ordered_Accs),
	group_iffs(Ordered_Accs, Loop, Iffs_Tmp),
	order_iffs_aux(Loop, Iffs_Tmp, Iffs_Out).
order_iffs([iff(Arg, Args) | Iffs], Accs, Pairs) :-
	term_variables(Args, Vars),
	order_iffs(Iffs, [Vars-Arg | Accs], Pairs).

order_iffs_aux(true, Iffs_In, Iffs_Out) :-
	deter_iffs(true, Iffs_In, Iffs_Out).
order_iffs_aux(false, Iffs, Iffs).

group_iffs([], false, []).
group_iffs([Vars-Arg | Pairs], Loop, Iffs) :-
	group_iffs(Pairs, Vars, Arg, Loop, Iffs).

group_iffs([], LVars, LArg, false, [iff(LArg, LVars)]).
group_iffs([Vars-Arg | Pairs], LVars, LArg, Loop, Iffs) :-
	Vars == LVars,
	!,
	Loop = true,
	Arg = LArg,
	group_iffs(Pairs, LVars, LArg, _, Iffs).
group_iffs([Vars-Arg | Pairs], LVars, LArg, Loop, [iff(LArg, LVars) | Iffs]) :-
	group_iffs(Pairs, Vars, Arg, Loop, Iffs).

deter_iffs(false, Iffs_In, Iffs_Out) :-
	order_iffs(Iffs_In, [], Iffs_Out).
deter_iffs(true, Iffs_In, Iffs_Out) :-
	deter_iffs(Iffs_In, [], Loop, Iffs_Tmp),
	deter_iffs(Loop, Iffs_Tmp, Iffs_Out).

deter_iffs([], Iffs_Acc, false, Iffs_Acc).
deter_iffs([iff(Arg, Args) | Iffs_In], Iffs_Acc, Loop, Iffs_Out) :-
	deter_iffs(Arg, Args, Iffs_In, Iffs_Acc, Loop, Iffs_Out).

deter_iffs(Arg, Args, Iffs_In, Iffs_Acc, Loop, Iffs_Out) :-
	Arg == true,
	!,
	true_args(Args),
	Loop = true,
	deter_iffs(Iffs_In, Iffs_Acc, _, Iffs_Out).
deter_iffs(Arg, Args, Iffs_In, Iffs_Acc, Loop, Iffs_Out) :-
	my_term_variable(Args, Arg),
	!,
	Loop = true,
	deter_iffs(Iffs_In, Iffs_Acc, _, Iffs_Out).
deter_iffs(Arg, Args, Iffs_In, Iffs_Acc, Loop, Iffs_Out) :-
	deter_iffs(Iffs_In, [iff(Arg, Args) | Iffs_Acc], Loop, Iffs_Out).

call_cond_assert(Goal, New_Goal) :- 
	\+ in_database(Goal, call),
%	format("~ncall(~w)", [Goal]),
	definite_assert(Goal, call, New_Goal).

answ_cond_assert(Goal, New_Goal) :- 
	\+ in_database(Goal, answ),
%	format("~nansw(~w)", [Goal]),
%	(ground(Goal) -> true; format("~n~w", [Goal])),
	definite_assert(Goal, answ, New_Goal).

definite_assert(Goal, Tag, New_Goal) :-
	functor(Goal, Name, Arity),
	functor(Old_Goal, Name, Arity),
	retract(fixpoint:user_atom(Old_Goal, Tag)), 
	!,
	term_subsumer(Goal, Old_Goal, New_Goal),
	asserta(fixpoint:user_atom(New_Goal, Tag)). 
definite_assert(Goal, Tag, Goal) :- 
	asserta(fixpoint:user_atom(Goal, Tag)).

true_args([]).
true_args([true | Args]) :-
	true_args(Args).

in_database(Goal, Tag) :- 
	Goal =.. [_ | Args],
	my_numbervars(Args, 1),
	fixpoint:user_atom(Goal, Tag),
	!.

my_ground([]).
my_ground([Arg | Args]) :-
	Arg == true,
	my_ground(Args).

my_numbervars([], _).
my_numbervars([Index | Args], Index) :-
	!,
	Index1 is Index + 1,
	my_numbervars(Args, Index1).
my_numbervars([_ | Args], Index) :-
	my_numbervars(Args, Index).

my_term_variable([], true).
my_term_variable([Arg | Args], Single_Var) :-
	Arg == true,
	!,
	my_term_variable(Args, Single_Var).
my_term_variable([Arg | Args], Arg) :-
	my_term_at_least_variable(Args, Arg).

my_term_at_least_variable([], _).
my_term_at_least_variable([Arg | Args], Single_Var) :-
	Arg == true,
	!,
	my_term_at_least_variable(Args, Single_Var).
my_term_at_least_variable([Arg | Args], Single_Var) :-
	Arg == Single_Var,
	my_term_at_least_variable(Args, Single_Var).
