:- module(pos_fixpoint, [td_con_fixpoint/1, td_msg_fixpoint/1, td_pos_fixpoint/1]).
:- use_module(library(terms)).
:- use_module(library(lists)).

td_con_fixpoint(Goal) :- td_fixpoint(Goal, con).
td_msg_fixpoint(Goal) :- td_fixpoint(Goal, msg).
td_pos_fixpoint(Goal) :- td_fixpoint(Goal, pos).

td_fixpoint(Goal, Type) :-	
	retractall(fixpoint:user_atom(_, _)),
        call_cond_assert(Goal, Type, Goal),
	(call_solve(Goal, Type); true),
	true.

call_solve(Head, Type) :-
        fixpoint:head_to_clause(Head, Body), 
        solve_right_atoms(Body, Type),
       	answ_cond_assert(Head, Type, Goal), 
        answ_solve(Goal, Type).

answ_solve(Atom, Type) :-
        fixpoint:atom_to_clause(Atom, Head, Left, Right), 
        fixpoint:user_atom(Head, call), 
        solve_left_atoms(Left, Type), 
        solve_right_atoms(Right, Type), 
        answ_cond_assert(Head, Type, Goal), 
        answ_solve(Goal, Type).

solve_left_atoms([], _).
solve_left_atoms([Atom | Atoms], Type) :- 
	functor(Atom, iff, _),
	!,
	solve_iff(Type, Atom),
	solve_left_atoms(Atoms, Type).
solve_left_atoms([fail | _], _) :- 
	!,
	fail.
solve_left_atoms([Atom | Atoms], Type) :- 
	fixpoint:user_atom(Atom, answ), 
	solve_left_atoms(Atoms, Type).

solve_right_atoms([], _).
solve_right_atoms([Atom | Atoms], Type) :- 
	functor(Atom, iff, _),
	!,
	solve_iff(Type, Atom),
	solve_right_atoms(Atoms, Type).
solve_right_atoms([fail | _], _) :- 
	!,
	fail.
solve_right_atoms([Atom | Atoms], Type) :-
        fixpoint:user_atom(Atom, answ),
	solve_right_atoms(Atoms, Type).
solve_right_atoms([Atom | _], Type) :-
        call_cond_assert(Atom, Type, Goal),
	call_solve(Goal, Type).

call_cond_assert(Goal, Type, New_Goal) :- 
	\+ in_database(Goal, call),
%	format("~ncall(~w)", [Goal]),
	definite_assert(Type, Goal, call, New_Goal).

answ_cond_assert(Goal, Type, New_Goal) :- 
	\+ in_database(Goal, answ),
%	format("~nansw(~w)", [Goal]),
	definite_assert(Type, Goal, answ, New_Goal).

definite_assert(con, Goal, Tag, New_Goal) :-
	functor(Goal, Name, Arity),
	functor(Old_Goal, Name, Arity),
	retract(fixpoint:user_atom(Old_Goal, Tag)), 
	!,
	Goal =.. [Name | Goal_Args],
	Old_Goal =.. [Name | Old_Args],
	con_args(Goal_Args, Old_Args, New_Args),
	New_Goal =.. [Name | New_Args],
	asserta(fixpoint:user_atom(New_Goal, Tag)). 
definite_assert(con, Goal, Tag, Goal) :- 
	asserta(fixpoint:user_atom(Goal, Tag)).

definite_assert(msg, Goal, Tag, New_Goal) :-
	functor(Goal, Name, Arity),
	functor(Old_Goal, Name, Arity),
	retract(fixpoint:user_atom(Old_Goal, Tag)), 
	!,
	term_subsumer(Goal, Old_Goal, New_Goal),
	asserta(fixpoint:user_atom(New_Goal, Tag)). 
definite_assert(msg, Goal, Tag, Goal) :- 
	asserta(fixpoint:user_atom(Goal, Tag)).

definite_assert(pos, Goal, Tag, Goal) :-
	asserta(fixpoint:user_atom(Goal, Tag)).

in_database(Goal, Tag) :- 
	Goal =.. [_ | Args],
	my_numbervars(Args, 1),
	fixpoint:user_atom(Goal, Tag),
	!.

solve_iff(con, Atom) :-
	Atom =.. [iff | Args],
	con_iff_args(Args).
solve_iff(msg, Atom) :-
	solve_iff(pos, Atom).
solve_iff(pos, iff(Arg)) :-
	!,
	Arg = true.
solve_iff(pos, iff(Arg_1, Arg_2)) :-
	!,
	Arg_1 = Arg_2.
solve_iff(pos, Atom) :-
	Atom =.. [iff | Args],
	pos_iff_args(Args).

con_iff_args([Arg | Args]) :-
	Arg == true,
	!,
	true_args(Args).
con_iff_args([Arg | Args]) :-
	ground_args(Args),
	!,
	Arg = true.
con_iff_args(_).

pos_iff_args(Args) :-
	true_args(Args).
pos_iff_args([false | Args]) :-
	member(false, Args).

true_args([]).
true_args([true | Args]) :-
	true_args(Args).

ground_args([]).
ground_args([Arg | Args]) :-
	Arg == true,
	ground_args(Args).

my_numbervars([], _).
my_numbervars([Index | Args], Index) :-
	!,
	Index1 is Index + 1,
	my_numbervars(Args, Index1).
my_numbervars([_ | Args], Index) :-
	my_numbervars(Args, Index).

con_args([], [], []).
con_args([Goal_Arg | Goal_Args], [Fact_Arg | Fact_Args], [Lub_Arg | Lub_Args]) :-
	con_arg(Goal_Arg, Fact_Arg, Lub_Arg),
	con_args(Goal_Args, Fact_Args, Lub_Args).

con_arg(Goal_Arg, Fact_Arg, Lub_Arg) :-
	Goal_Arg == true,
	Fact_Arg == true,
	!,
	Lub_Arg = true.
con_arg(_, _, _).
