:- module(pretty, [pretty_print/2]).

pretty_print(con_nonground_model, OStream) :- 	
	pretty_nonground_model_print(OStream).
pretty_print(det_nonground_model, OStream) :- 	
	pretty_nonground_model_print(OStream).
pretty_print(msg_nonground_model, OStream) :- 	
	pretty_nonground_model_print(OStream).
pretty_print(imp_nonground_model, OStream) :- 	
	pretty_nonground_model_print(OStream).
pretty_print(def_nonground_model, OStream) :- 	
	pretty_nonground_model_print(OStream).
pretty_print(pos_nonground_model, OStream) :- 	
	pretty_nonground_model_print(OStream).
pretty_print(sym_nonground_model, OStream) :- 	
	pretty_symbolic_model_print(OStream).

pretty_nonground_model_print(OStream) :-
	fixpoint:user_atom(Atom, Tag), 
	check_model_exit(Tag, Atom),
	my_portray_clause(OStream, Tag, Atom),
	fail.
pretty_nonground_model_print(OStream) :-
	portray_clause(OStream, end_of_output).

pretty_symbolic_model_print(OStream) :-
	fixpoint:user_atom(Atom, _, _, Tag), 
	my_portray_clause(OStream, Tag, Atom),
	fail.
pretty_symbolic_model_print(OStream) :-
	portray_clause(OStream, end_of_output).

check_model_exit(call, Atom) :-
	functor(Atom, Name, Arity),
	functor(New_Atom, Name, Arity),
	fixpoint:user_atom(New_Atom, answ),
	!.
check_model_exit(call, Atom) :-
	functor(Atom, Name, Arity),
	format("{warning, ~w/~w goals cannot succeed}~n", [Name, Arity]).
check_model_exit(answ, _).

my_portray_clause(OStream, Tag, Atom) :-
	functor(Atom, Name, 0),
	!,
	format(OStream, "~w(", [Tag]),
	portray_name(OStream, Name),
	format(OStream, ").~n", []).
my_portray_clause(OStream, Tag, Atom) :-
	Atom =.. [Name | Args],
	!,
	format(OStream, "~w(", [Tag]),
	portray_name(OStream, Name),
	format(OStream, "(", []),
	portray_args(Args, OStream),
	format(OStream, ")).~n", []).

portray_name(OStream, Name) :-
	name(Name, [C | Cs]),
	97 =< C, C =< 122,
	alpha_numeric_string(Cs),
	!,
	format(OStream, "~w", [Name]).
portray_name(OStream, Name) :-
	format(OStream, "'~w'", [Name]).

alpha_numeric_string([]).
alpha_numeric_string([C | Cs]) :-
	97 =< C, C =< 122, !,
	alpha_numeric_string(Cs).
alpha_numeric_string([C | Cs]) :-
	65 =< C, C =< 90, !,
	alpha_numeric_string(Cs).
alpha_numeric_string([C | Cs]) :-
	48 =< C, C =< 57, !,
	alpha_numeric_string(Cs).
alpha_numeric_string([95 | Cs]) :-
	alpha_numeric_string(Cs).

portray_args([Arg], OStream) :-
	format(OStream, "~w", [Arg]).
portray_args([Arg1, Arg2 | Args], OStream) :-
	format(OStream, "~w, ", [Arg1]),
	portray_args([Arg2 | Args], OStream).

