:- module(pure, [purify/2]).

purify(_, OStream) :-
	fixpoint:head_to_clause(Head, Body),
	pure_atom(Head, Pure_Head),
	pure_atoms(Body, Pure_Body),
	unflat_atoms(Pure_Body, Unflat_Body),
	portray_clause(OStream, (Pure_Head :- Unflat_Body)),
	fail.
purify(Prog, OStream) :-
	benchmarks:benchmark(Prog, _, _, _, Goal),
	portray_clause(OStream, (:- Goal)).

pure_atoms([], []).
pure_atoms([Atom | Atoms], [Pure_Atom | Pure_Atoms]) :-
	pure_atom(Atom, Pure_Atom),
	pure_atoms(Atoms, Pure_Atoms).

pure_atom(iff(Arg), Pure_Atom) :-
	!,
	Pure_Atom =.. [=, Arg, c].
pure_atom(iff(Left_Arg, Right_Arg), Pure_Atom) :-
	!,
	Pure_Atom =.. [=, Left_Arg, Right_Arg].
pure_atom(Atom, Pure_Atom) :-
	Atom =.. [iff, Left_Arg | Right_Args],
	!,
	Right_Arg =.. [f | Right_Args],
	Pure_Atom =.. [=, Left_Arg, Right_Arg].
pure_atom(fail, Pure_Atom) :-
	!,
	Pure_Atom = fail.
pure_atom(Atom, Pure_Atom) :-	
	Atom =.. [Name | Args],
	pure_args(Args, Pure_Args),
	Pure_Atom =.. [Name | Pure_Args].

pure_args([], []).
pure_args([Arg | Args], [Pure_Arg | Pure_Args]) :-
	Arg == true,
	!,
	Pure_Arg = c,
	pure_args(Args, Pure_Args).
pure_args([Arg | Args], [Arg | Pure_Args]) :-
	pure_args(Args, Pure_Args).

unflat_atoms([], true).
unflat_atoms([Atom], Unflat_Atoms) :-
	!,
	Unflat_Atoms = Atom.
unflat_atoms([Atom | Atoms], (Atom, Unflat_Atoms)) :-
	unflat_atoms(Atoms, Unflat_Atoms).
